from itertools import islice, tee


def is_sorted(list, /, comparator=None, key=lambda x: x):
    if comparator is None:

        def comparator(lhs, rhs):
            return key(lhs) <= key(rhs)

    l1, l2 = tee(list, 2)
    l2 = islice(l2, 1, None)
    return all(comparator(r1, r2) for r1, r2 in zip(l1, l2))


def is_unique(list, /, key=lambda x: x):
    res = set()
    for x in list:
        k = key(x)
        if k in res:
            return False
        res.add(k)
    return True


def find_duplicates(list, /, key=lambda x: x):
    res = set()
    duplicates = set()
    for x in list:
        k = key(x)
        if k in res:
            if k not in duplicates:
                yield k
                duplicates.add(k)
        else:
            res.add(k)


def group(list, /, key=lambda x: x):
    groups = {}
    for x in list:
        k = key(x)
        records = groups.setdefault(k, [])
        records.append(x)
    return groups


def remap_keys(data, mapping):
    if isinstance(mapping, dict):
        mapping = mapping.items()
    return {new: data[old] for old, new in mapping if old in data}


def extract_keys(data, keys):
    return {k: data.get(k) for k in keys}



import json
import logging

import click
from cloudflare import Cloudflare

log = logging.getLogger("Compliance check")


@click.command("list", help="List Cloudflare Zones")
@click.option(
    "-t",
    "--token",
    type=str,
    # The lib automaticly consider this option
    # envvar="CLOUDFLARE_API_TOKEN",
    help="Cloudflare API token",
)
@click.option(
    "-j",
    "--json",
    "json_format",
    type=str,
    is_flag=True,
    default=False,
    help="Cloudflare API token",
)
def list_zones(token, json_format):
    cf_config = {}
    if token:
        cf_config["token"] = token
    cf = Cloudflare(**cf_config)
    zones = [z.model_dump() for z in cf.zones.list()]
    if json_format:
        print(json.dumps(zones, indent=4))
    else:
        for z in zones:
            print("{name} (ID: {id})".format(**z))



import logging

import click
from nagra_network_misc_utils.logger import set_default_logger

from nagra_network_cloudflare_utils.compliance import remove_cloudflare_records
from nagra_network_cloudflare_utils.dns_checker import check_csvfile, sort_csvfile
from nagra_network_cloudflare_utils.list_zones import list_zones

set_default_logger()
logging.getLogger().setLevel(logging.WARNING)


LOG_LEVELS = {
    "critical": logging.CRITICAL,
    "fatal": logging.FATAL,
    "error": logging.ERROR,
    "warn": logging.WARNING,
    "warning": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
    "notset": logging.NOTSET,
}


# Keep it as the main entry point
@click.group()
@click.option(
    "--log-level",
    type=click.Choice(
        ["critical", "fatal", "error", "warn", "warning", "info", "debug", "notset"]
    ),
    default="info",
)
def main(log_level):
    # This sets a convenient global handler for the logs
    set_default_logger()
    logging.getLogger().setLevel(LOG_LEVELS[log_level])


main.add_command(check_csvfile)
main.add_command(sort_csvfile)
main.add_command(remove_cloudflare_records)
main.add_command(list_zones)

if __name__ == "__main__":
    main()



import logging
from typing import Iterable, Literal, Optional, Union

from nagra_network_misc_utils.gitlab import add_mr_comment
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PositiveInt,
    StrictBool,
    TypeAdapter,
    field_validator,
)

from .schema_utils import group, is_sorted

log = logging.getLogger("Validation")

PROXIED_VALUES = Literal["false", "true"]
TYPES_VALUES = Literal[
    "A",
    "AAAA",
    "CNAME",
    "MX",
    "TXT",
    "CAA",
    "SRV",
    "PTR",
    "SOA",
    "NS",
    "DS",
    "DNSKEY",
    "LOC",
    "NAPTR",
    "SSHFP",
    "SVCB",
    "TSLA",
    "URI",
    "SPF",
]


# Same schema to validate tfplan.json, Cloudflare output and the csv file
class Record(BaseModel):
    # https://docs.pydantic.dev/latest/api/config/#pydantic.config.ConfigDict.populate_by_name
    model_config = ConfigDict(populate_by_name=True)

    id: Optional[str] = None
    name: str
    type: TYPES_VALUES
    value: str = Field(alias="content")
    ttl: PositiveInt
    proxied: Union[StrictBool, PROXIED_VALUES]

    @field_validator("name")
    def no_trailing_dot(cls, value):
        if value.strip().endswith("."):
            raise ValueError('dns entry are not allowed to end with a "."')
        return value

    def get_uuid(self):
        return (self.name, self.value)

    def __str__(self):
        return f"{self.name}, {self.type}, {self.value}"


RecordList = TypeAdapter(list[Record])


def check_ns_records(records: Iterable[Record], report_on_mr=True) -> bool:
    apex_ns_records = []
    delegation_ns_records = set()
    other_records: list[Record] = []
    for rec in records:
        if rec.type != "NS":
            other_records.append(rec)
        # We must ignore APEX NS records
        elif not rec.name or rec.name in ("", "@"):
            apex_ns_records.append(rec)
        else:
            delegation_ns_records.add(rec.name)

    invalid_records: dict[str, list[Record]] = {}
    for deleg in delegation_ns_records:
        sub_deleg = f".{deleg}"
        for rec in other_records:
            if rec.name == deleg or rec.name.endswith(sub_deleg):
                g = invalid_records.setdefault(deleg, [])
                g.append(rec)

    if not invalid_records:
        return True

    error_list = "\n".join(
        f"- {r.name} ({r.type}) => {deleg}"
        for deleg, records in invalid_records.items()
        for r in records
    )
    error_msg = f"""\
The following records are under a delegated zone (i.e. the records ends with the same value as an existing NS record)
The record is therefore not valid and must be moved to the correct zone.

{error_list}
"""
    log.error(error_msg)
    if report_on_mr:
        add_mr_comment(error_msg)
    return False


def check_duplicates(records: Iterable[Record], report_on_mr=True) -> bool:
    grouped = group(records, lambda r: r.name)
    grouped = {k: v for k, v in grouped.items() if len(v) > 1}
    if not grouped:
        return True
    duplicates = []
    for name, records in grouped.items():
        # Some entries can be duplicated
        types = {r.type for r in records}
        types -= {"MX", "TXT"}  # They can be added along with any other
        # A and AAAA records are compatible
        if any(t in types for t in ("A", "AAAA")):
            types -= {"A", "AAAA"}
            if types:
                duplicates.append(name)
        # We can have multiple value for NS records
        if "NS" in types and len(types) > 1:
            duplicates.append(name)
    if not duplicates:
        return True
    log.warning(
        ("There are duplicate entries," "be sure that it is what you want:\n{}").format(
            "\n".join(grouped.keys())
        )
    )
    duplicates_str = "\n".join(duplicates)
    error_msg = f"""\
The following records have duplicates
Note that some records can be duplicated (e.g. A, AAAA, NS),
and some types are compatibles (e.g. A and AAAA):
{duplicates_str}
"""
    if report_on_mr:
        add_mr_comment(error_msg)
    return False


def check_records(records, raise_on_invalid=True, report_on_mr=True):
    records: Iterable[Record] = RecordList.validate_python(records)  # model_validate
    valid = True
    valid &= check_duplicates(records, report_on_mr=report_on_mr)
    valid &= check_ns_records(records, report_on_mr=report_on_mr)
    if not is_sorted(records, key=lambda x: x.name):
        # NOTE: This must not fail the pipeline. This is an additional warning
        # The error is located where the data are defined.
        log.warning("Records are not sorted, please sort them")
    if not valid and raise_on_invalid:
        raise Exception("Records are invalids. Check the logs for more information.")
    return records, valid


def sort_records(records):
    # NOTE: `sorted` is stable: https://stackoverflow.com/questions/1915376/is-pythons-sorted-function-guaranteed-to-be-stable
    # meaning: in case of equality, the order of the output is the same as the input
    yield from sorted(records, key=lambda r: (r["name"], r["type"]))



import csv
import logging
from io import StringIO
from pathlib import Path

import click
from nagra_network_misc_utils.gitlab import add_mr_comment, post_git_diff

from .schema import check_records, sort_records

log = logging.getLogger("Record checker")


@click.command(
    "check",
    help="Check that the CSV file is compliant with Terraform module",
)
@click.option(
    "-f",
    "--file",
    type=Path,
    help="Name of the file to check",
    default="dns_records.csv",
)
def check_csvfile(file: Path):
    if not file.is_file():
        log.warn("File does not exist. Ignoring validation")
        return
    if file.suffix != ".csv":
        log.warn("File must be a .csv file")
        return
    with open(file) as f:
        try:
            check_records(csv.DictReader(f))
        except Exception as e:
            log.error(e)
            add_mr_comment(str(e))
            exit(1)
    log.info(f"File {file} is valid")


@click.command(
    "sort",
    help="Sort the CSV file",
)
@click.option(
    "-f",
    "--file",
    type=Path,
    help="Name of the file to check",
    default="dns_records.csv",
)
def sort_csvfile(file: Path):
    header = ("name", "type", "content", "ttl", "proxied")
    content = file.read_text()

    # Sort file
    records = list(csv.DictReader(StringIO(content), fieldnames=header))
    records = records[1:]  # Remove header
    records = sort_records(records)
    output_buffer = StringIO()
    writer = csv.DictWriter(
        output_buffer,
        fieldnames=header,
        lineterminator="\n",
    )
    writer.writeheader()
    writer.writerows(records)
    result = output_buffer.getvalue()

    # Dump file
    file.write_text(result)
    # Report difference in git if needed
    post_git_diff()
    # Exit with error status code if a difference is detected
    if result != content:
        logging.info("File was not correctly sorted")
        exit(1)



import json
import logging
from pathlib import Path

import click
from cloudflare import Cloudflare
from glom import flatten, glom

from .schema import check_records

log = logging.getLogger("Compliance check")


def check_records_compliance(records_from_cloudflare, records_from_file):
    for record in records_from_file:
        if record in records_from_cloudflare:
            records_from_cloudflare.remove(record)
    return records_from_cloudflare


def extract_records_dict(data):
    return [
        r["value"]
        for r in flatten(glom(data, ("**.resources")))
        if r["type"] == "cloudflare_record" and "value" in r
    ]


def records_from_tf_output(file):
    with open(file) as f:
        data = json.load(f)
    values = extract_records_dict(data)
    records, _valid = check_records(values)
    return {r.id: r for r in records}


def records_from_cloudflare(cf, zone_id):
    cloudflare_records = [
        rec.model_dump() for rec in cf.dns.records.list(zone_id=zone_id, per_page=1000)
    ]
    cloudflare_records, _valid = check_records(cloudflare_records)
    return {r.id: r for r in cloudflare_records}


@click.command(
    "remove", help="Remove Cloudflare DNS records that are not defined in Terraform"
)
@click.option(
    "-t",
    "--token",
    type=str,
    # The lib automaticly consider this option
    # envvar="CLOUDFLARE_API_TOKEN",
    help="Cloudflare API token",
)
@click.option(
    "-z",
    "--zone",
    "zone_id",
    type=str,
    envvar="ZONE_ID",
    help="Cloudflare Zone to process",
)
@click.option(
    "-f",
    "--file",
    "statefile",
    type=Path,
    help="Terraform tfplan file (must be in json format)",
    default="plan.tfplan.json",
)
@click.option(
    "--test",
    type=bool,
    is_flag=True,
    help=("Do not delete, only display the records " "that would have been deleted"),
    default=False,
)
def remove_cloudflare_records(token, zone_id, statefile, test):
    cf_config = {}
    if token:
        cf_config["token"] = token
    cf = Cloudflare(**cf_config)
    cf_records = records_from_cloudflare(cf, zone_id)
    state_records = records_from_tf_output(statefile)
    records_to_remove = [
        r for r_id, r in cf_records.items() if r_id not in state_records
    ]
    records_to_remove.sort(key=lambda r: r.name)
    if not records_to_remove:
        log.warn("No record to remove")
    elif test:
        print(
            "The following records should be removed:\n{}".format(
                "\n".join(str(r) for r in records_to_remove)
            )
        )
    else:
        log.warn(
            "The following records will be removed:\n{}".format(
                "\n".join(str(r) for r in records_to_remove)
            )
        )
        for record in records_to_remove:
            cf.remove(zone_id, record.id)
        log.info("Records have been removed")
